"""Helper functions for documentation."""

from __future__ import annotations

import functools
import inspect

from typing import Any, Callable, Dict, Protocol, TypeVar, Union, cast

from typing_extensions import TypeAlias

F = TypeVar('F', bound=Callable[..., Any])

MetaDocsParams: TypeAlias = Dict[str, Dict[str, str]]
MetaDocs: TypeAlias = Dict[str, Union[str, MetaDocsParams]]


class DecoratedMetaDocsFunction(Protocol):
    """Type for functions with _meta_docs."""

    _meta_docs: MetaDocs

    def __call__(self, *args: Any, **kwargs: Any) -> Any:
        """Define a placeholder for the type."""
        ...


def docparams(
    param_docs: Dict[str, str],
) -> Callable[[F], DecoratedMetaDocsFunction]:
    """
    Decorate functions to add dynamic parameters to docs and metadata.

    Decorator to update the function's docstring with parameter documentation
    and add a '_meta_docs' attribute containing parameter details.

    Parameters
    ----------
    param_docs : Dict[str, str]
        A dictionary mapping parameter names to their docstring descriptions.

    Returns
    -------
    Callable
        The decorated function with updated docstring and '_meta_docs'.
    """

    def decorator(func: F) -> DecoratedMetaDocsFunction:  # noqa: PLR0912
        @functools.wraps(func)
        def wrapper(*args: Any, **kwargs: Any) -> Any:
            return func(*args, **kwargs)

        sig = inspect.signature(func)
        annotations = func.__annotations__

        # Build the _meta_docs dictionary
        parameters: MetaDocsParams = {}

        for name, param in sig.parameters.items():
            if name == 'self':
                continue
            param_info = {}
            if name in annotations:
                param_type = annotations[name]
                if isinstance(param_type, type):
                    param_info['type'] = param_type.__name__
                else:
                    param_info['type'] = str(param_type)
            if name in param_docs:
                param_info['help'] = param_docs[name]
            if param.default != inspect.Parameter.empty:
                param_info['default'] = param.default
            parameters[name] = param_info

        # Build parameter documentation
        param_docs_lines = []
        for name, info in parameters.items():
            param_line = f'{name} : {info.get("type", "")}'
            param_docs_lines.append(param_line)
            default = info.get('default', None)
            if default is not None:
                param_docs_lines.append(f'    Default: {default}')
            docstring = info.get('help', '')
            if docstring:
                param_docs_lines.append(f'    {docstring}')
            param_docs_lines.append('')  # Empty line for separation

        # Update the function's docstring
        doc = func.__doc__ or ''
        # Split the docstring into lines
        doc_lines = doc.strip().split('\n')
        # Find the 'Parameters' section
        params_index = None
        for i, line in enumerate(doc_lines):
            if line.strip().lower() in ('parameters:', 'parameters'):
                params_index = i
                break
        if params_index is not None:
            # Remove existing parameter documentation
            next_section_index = None
            for j in range(params_index + 1, len(doc_lines)):
                if doc_lines[j].strip() == '':
                    continue
                elif doc_lines[j].startswith(' '):
                    continue  # Indented parameter documentation
                elif doc_lines[j].endswith(':') and not doc_lines[
                    j
                ].startswith(' '):
                    next_section_index = j
                    break
            if next_section_index is None:
                next_section_index = len(doc_lines)
            doc_lines = (
                doc_lines[: params_index + 1] + doc_lines[next_section_index:]
            )
            doc_lines.insert(params_index + 1, '')
            doc_lines[params_index + 2 : params_index + 2] = param_docs_lines
        else:
            # Append 'Parameters' section at the end
            if doc_lines and doc_lines[-1].strip() != '':
                doc_lines.append('')
            doc_lines.append('Parameters:')
            doc_lines.append('-----------')
            doc_lines.append('')
            doc_lines.extend(param_docs_lines)

        # Reconstruct the docstring
        updated_doc = '\n'.join(doc_lines)
        wrapper.__doc__ = updated_doc

        meta_docs: MetaDocs = {
            'title': func.__doc__ or '',
            'parameters': parameters,
        }

        # Set the _meta_docs attribute
        setattr(wrapper, '_meta_docs', meta_docs)

        return cast(DecoratedMetaDocsFunction, wrapper)

    return decorator



"""SugarBase classes for containers."""

from __future__ import annotations

import io
import os
import shlex
import sys

from copy import deepcopy
from pathlib import Path
from typing import Any, Union

import dotenv
import sh
import yaml

from jinja2 import Environment

from sugar import __version__
from sugar.logs import SugarErrorType, SugarLogs

TEMPLATE = Environment(
    autoescape=False,
    variable_start_string='${{',
    variable_end_string='}}',
)


class SugarBase:
    """SugarBase defined the base structure for the Sugar classes."""

    actions: list[str] = []
    args: dict[str, str] = {}
    file: str = ''
    config: dict[str, Any] = {}
    # note: it starts with a simple command
    #       it is replaced later in the execution
    backend_app: sh.Command = sh.echo
    backend_args: list[str] = []
    defaults: dict[str, Any] = {}
    dry_run: bool = False
    env: dict[str, str] = {}
    options_args: list[str] = []
    cmd_args: list[str] = []
    service_group: dict[str, Any] = {}
    service_names: list[str] = []
    group_selected: str = ''
    verbose: bool = False

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Initialize the actions list for all the created commands."""
        super().__init_subclass__(**kwargs)
        # Ensure each subclass has its own actions list
        cls.actions = cls.actions.copy()
        prefix = '_cmd_'
        prefix_len = len(prefix)
        for name, value in cls.__dict__.items():
            if callable(value) and name.startswith(prefix):
                action_name = name[prefix_len:]
                cls.actions.append(action_name)

    def __init__(self) -> None:
        """Initialize SugarBase instance."""
        self.file = '.sugar.yaml'
        self.verbose = False
        self.dry_run = False
        self.args: dict[str, str] = {}
        self.options_args: list[str] = []
        self.cmd_args: list[str] = []
        self.config: dict[str, Any] = {}
        self.backend_args: list[str] = []
        self.defaults: dict[str, Any] = {}
        self.env: dict[str, str] = {}
        self.service_group: dict[str, Any] = {}
        self.service_names: list[str] = []
        self.group_selected: str = ''

    def _setup_load(self, **kwargs: Any) -> None:
        """Set up the configuration for running the commands."""
        args: dict[str, str] = kwargs.get('args', {})
        options_args: list[str] = kwargs.get('options_args', [])
        cmd_args: list[str] = kwargs.get('cmd_args', [])

        self.args = deepcopy(args)
        self.options_args = deepcopy(options_args)
        # note: it needs to be refactored
        self.cmd_args = deepcopy(cmd_args)

        self._load_config()
        self._load_env()
        self._load_defaults()
        self._load_root_services()

        self._load_backend_app()
        self._load_backend_args()

        self._verify_config()

    def _call_backend_app(
        self,
        action: str,
        services: list[str] = [],
        options_args: list[str] = [],
        cmd_args: list[str] = [],
        _out: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stdout,
        _err: Union[io.TextIOWrapper, io.StringIO, Any] = sys.stderr,
    ) -> None:
        sh_extras = {
            '_in': sys.stdin,
            '_out': _out,
            '_err': _err,
            '_no_err': True,
            '_env': os.environ,
            '_bg': True,
            '_bg_exc': False,
        }

        positional_parameters = [
            *self.backend_args,
            *[action],
            *options_args,
            *services,
            *cmd_args,
        ]

        if self.verbose or self.dry_run:
            SugarLogs.print_info(
                f'>>> {self.backend_app} {" ".join(positional_parameters)}'
            )
            SugarLogs.print_info('-' * 80)

        if self.dry_run:
            SugarLogs.print_warning(
                'Running it in dry-run mode, the command was skipped.'
            )
            return

        p = self.backend_app(
            *positional_parameters,
            **sh_extras,
        )

        try:
            p.wait()
        except sh.ErrorReturnCode as e:
            SugarLogs.raise_error(str(e), SugarErrorType.SH_ERROR_RETURN_CODE)
        except KeyboardInterrupt:
            pid = p.pid
            p.kill()
            SugarLogs.raise_error(
                f'Process {pid} killed.', SugarErrorType.SH_KEYBOARD_INTERRUPT
            )

    def _check_config_file(self) -> bool:
        return Path(self.file).exists()

    # Check if services item is given
    def _check_services_item(self) -> bool:
        return hasattr(self.config, 'services')

    # set default group main
    def _load_root_services(self) -> None:
        """Load services attribute in the root of the configuration."""
        # must set the default group
        services = self.config.get('services', {})

        if not services:
            return

        self.config['groups'] = {
            'main': {
                'project-name': services.get('project-name'),
                'config-path': services.get('config-path'),
                'env-file': services.get('env-file'),
                'services': {
                    'default': services.get('default'),
                    'available': services.get('available'),
                },
            }
        }
        self.defaults['group'] = 'main'
        self.group_selected = 'main'
        self.service_group = deepcopy(self.config['groups']['main'])
        del self.config['services']

    def _filter_service_group(self) -> None:
        groups = self.config['groups']

        if not self.group_selected:
            default_group = self.defaults.get('group')
            if not default_group:
                SugarLogs.raise_error(
                    'The service group parameter or default '
                    "group configuration weren't defined.",
                    SugarErrorType.SUGAR_INVALID_PARAMETER,
                )
            selected_group_name = default_group
        else:
            selected_group_name = self.group_selected

        # Verify if project-name is not null
        default_project_name = self.defaults.get('project-name', '') or ''

        for group_name, group_data in groups.items():
            if group_name == selected_group_name:
                if default_project_name and 'project-name' not in group_data:
                    # just use default value if "project-name" is not set
                    group_data['project-name'] = default_project_name
                if not group_data.get('services', {}).get('default'):
                    # if default is not given or it is empty or null,
                    # use as default all the services available
                    default_services = [
                        service['name']
                        for service in group_data.get('services', {}).get(
                            'available'
                        )
                    ]
                    group_data['services']['default'] = ','.join(
                        default_services
                    )
                self.service_group = group_data
                return

        SugarLogs.raise_error(
            f'The given group service "{group_name}" was not found '
            'in the configuration file.',
            SugarErrorType.SUGAR_MISSING_PARAMETER,
        )

    def _load_config(self) -> None:
        with open(self.file, 'r') as f:
            # escape template tags
            content = f.read()
            f_content = io.StringIO(content)
            self.config = yaml.safe_load(f_content)

        # check if either  services or  groups are present
        if not (self.config.get('services') or self.config.get('groups')):
            SugarLogs.raise_error(
                'Either `services` OR  `groups` flag must be given',
                SugarErrorType.SUGAR_INVALID_CONFIGURATION,
            )
        # check if both services and groups are present
        if self.config.get('services') and self.config.get('groups'):
            SugarLogs.raise_error(
                '`services` and `groups` flags given, only 1 is allowed.',
                SugarErrorType.SUGAR_INVALID_CONFIGURATION,
            )

    def _load_backend_app(self) -> None:
        backend_cmd = self.config.get('backend', '')
        supported_backends = ['compose']

        if backend_cmd not in supported_backends:
            SugarLogs.raise_error(
                f'"{self.config["backend"]}" not supported yet.'
                f' Supported backends are: {", ".join(supported_backends)}.',
                SugarErrorType.SUGAR_COMPOSE_APP_NOT_SUPPORTED,
            )

        self.backend_app = sh.docker
        self.backend_args.append(backend_cmd)

    def _load_backend_args(self) -> None:
        self._filter_service_group()

        if self.service_group.get('env-file'):
            self.backend_args.extend(
                ['--env-file', self.service_group['env-file']]
            )

        config_path = []
        backend_path_arg = self.service_group['config-path']
        if isinstance(backend_path_arg, str):
            config_path.append(backend_path_arg)
        elif isinstance(backend_path_arg, list):
            config_path.extend(backend_path_arg)
        else:
            SugarLogs.raise_error(
                'The attribute config-path` just supports the data '
                f'types `string` or `list`, {type(backend_path_arg)} '
                'received.',
                SugarErrorType.SUGAR_INVALID_CONFIGURATION,
            )

        for p in config_path:
            self.backend_args.extend(['--file', p])

        if self.service_group.get('project-name'):
            self.backend_args.extend(
                ['--project-name', self.service_group['project-name']]
            )

    def _load_defaults(self) -> None:
        _defaults = self.config.get('defaults', {})

        for k, v in _defaults.items():
            unescaped_value = v if isinstance(v, str) else str(v)

            _defaults[k] = yaml.safe_load(
                TEMPLATE.from_string(unescaped_value).render(env=self.env)
            )

        self.defaults = _defaults

    def _load_env(self) -> None:
        self.env = dict(os.environ)

        env_file = self.config.get('env-file', '')

        if not env_file:
            return

        if not env_file.startswith('/'):
            # use .sugar file as reference for the working
            # directory for the .env file
            env_file = str(Path(self.file).parent / env_file)

        if not Path(env_file).exists():
            SugarLogs.raise_error(
                'The given env-file was not found.',
                SugarErrorType.SUGAR_INVALID_CONFIGURATION,
            )
        self.env.update(dotenv.dotenv_values(env_file))  # type: ignore

    def _get_list_args(self, args: str) -> list[str]:
        """Return a list with the name of the service if any."""
        if not args.strip():
            return []

        return shlex.split(args)

    def _get_service_name(self, service: str) -> list[str]:
        """Return a list with the name of the service if any."""
        return [service] if service else []

    def _get_services_names(self, **kwargs: Any) -> list[str]:
        if 'all' not in kwargs and 'services' not in kwargs:
            # the command doesn't specify services (e.g. version)
            return []

        _arg_services = kwargs.get('services', '')
        _arg_all = kwargs.get('all', '')

        services_config = self.service_group['services']
        service_names: list[str] = []
        services_default = services_config.get('default', '')

        if _arg_all:
            service_names = [
                v['name']
                for v in self.service_group.get('services', {}).get(
                    'available'
                )
            ]
        elif _arg_services == '' and not services_default:
            SugarLogs.raise_error(
                'If you want to execute the operation for all services, '
                'use --all parameter.',
                SugarErrorType.SUGAR_INVALID_PARAMETER,
            )
        elif _arg_services:
            service_names = _arg_services.split(',')
        elif services_config.get('default'):
            service_names = services_default.split(',')

        return service_names

    def _verify_config(self) -> None:
        if not self._check_config_file():
            SugarLogs.raise_error(
                'Config file .sugar.yaml not found.',
                SugarErrorType.SUGAR_INVALID_CONFIGURATION,
            )

        if not len(self.config['groups']):
            SugarLogs.raise_error(
                'No service groups found.',
                SugarErrorType.SUGAR_INVALID_CONFIGURATION,
            )

    def _version(self) -> None:
        SugarLogs.print_info(f'Sugar Version: {__version__}')

    def load(
        self,
        file: str = '.sugar.yaml',
        group: str = '',
        dry_run: bool = False,
        verbose: bool = False,
    ) -> None:
        """Load makim configuration."""
        self.file = file
        self.group_selected = group
        self.dry_run = dry_run
        self.verbose = verbose

        self._setup_load()



"""Functions about console."""

from __future__ import annotations

import os


def get_terminal_size() -> tuple[int, int]:
    """Return the height (number of lines) of the terminal using os module."""
    size = os.get_terminal_size()
    try:
        height = size.lines
    except OSError:
        # Default to 24 lines if the terminal size cannot be determined.
        height = 24

    try:
        width = size.columns
    except OSError:
        # Default to 24 lines if the terminal size cannot be determined.
        height = 80
    return width, height



"""Logs classes and function for sugar system."""

from __future__ import annotations

import os

from enum import Enum

from colorama import Fore


class SugarErrorType(Enum):
    """SugarErrorType group all error types handled by the system."""

    SH_ERROR_RETURN_CODE = 1
    SH_KEYBOARD_INTERRUPT = 2
    SUGAR_COMPOSE_APP_NOT_SUPPORTED = 3
    SUGAR_COMPOSE_APP_NOT_FOUNDED = 4
    SUGAR_INVALID_PARAMETER = 5
    SUGAR_MISSING_PARAMETER = 6
    SUGAR_INVALID_CONFIGURATION = 7
    SUGAR_ACTION_NOT_IMPLEMENTED = 8
    SUGAR_NO_SERVICES_RUNNING = 9


class SugarLogs:
    """SugarLogs is responsible for handling system messages."""

    @staticmethod
    def raise_error(
        message: str,
        message_type: SugarErrorType = SugarErrorType.SH_ERROR_RETURN_CODE,
    ) -> None:
        """Print error message and exit with given error code."""
        print(Fore.RED, f'[EE] {message}', Fore.RESET)
        os._exit(message_type.value)

    @staticmethod
    def print_info(message: str) -> None:
        """Print info message."""
        print(Fore.BLUE, message, Fore.RESET)

    @staticmethod
    def print_warning(message: str) -> None:
        """Print warning message."""
        print(Fore.YELLOW, message, Fore.RESET)



"""Definition of the CLI structure."""

from __future__ import annotations

import os
import sys

from pathlib import Path
from typing import Any, Callable, Dict, Optional, Type, Union, cast

import click
import typer

from typer import Option

from sugar import __version__
from sugar.core import extensions
from sugar.docs import MetaDocs, MetaDocsParams
from sugar.logs import SugarLogs

# "count" means the number of parameters expected for each flag
CLI_ROOT_FLAGS_VALUES_COUNT = {
    '--dry-run': 0,
    '--file': 1,
    '--group': 1,
    '--help': 0,  # not necessary to store this value
    '--verbose': 0,
    '--version': 0,  # not necessary to store this value
}

flags_state: dict[str, bool] = {
    'verbose': False,
}

flags_group: dict[str, str] = {
    'group': '',
}

flags_dry_run: dict[str, bool] = {
    'dry_run': False,
}

opt_state: dict[str, list[Any]] = {
    'options': [],
    'cmd': [],
}

sugar_exts = {
    ext_name: ext_class() for ext_name, ext_class in extensions.items()
}

typer_groups: dict[str, typer.Typer] = {}

app = typer.Typer(
    name='sugar',
    help=(
        'Sugar is a tool that help you to organize'
        "and simplify your containers' stack."
    ),
    epilog=(
        'If you have any problem, open an issue at: '
        'https://github.com/osl-incubator/sugar'
    ),
    short_help="Sugar is a tool that help you \
        to organize containers' stack",
)


def _check_sugar_file(file_path: str = '.sugar.yaml') -> bool:
    return Path(file_path).exists()


def version_callback() -> None:
    """Print the Sugar version."""
    SugarLogs.print_info(f'Sugar version: {__version__}')


@app.callback(invoke_without_command=True)
def main(
    ctx: typer.Context,
    group: str = Option(
        '',
        '--group',
        help='Set the group of services for running the sugar command.',
    ),
    version: bool = Option(
        None,
        '--version',
        '-v',
        is_flag=True,
        is_eager=True,
        help='Show the version of sugar.',
    ),
    verbose: bool = Option(
        False,
        '--verbose',
        is_flag=True,
        is_eager=True,
        help='Show the command executed.',
    ),
    dry_run: bool = Option(
        False,
        '--dry-run',
        is_flag=True,
        is_eager=True,
        help="Don't actually execute the command.",
    ),
) -> None:
    """
    Process commands for specific flags.

    Otherwise, show the help menu.
    """
    ctx.ensure_object(dict)

    if verbose:
        # global
        flags_state['verbose'] = True

    if group:
        # global
        flags_group['group'] = group

    if dry_run:
        # global
        flags_dry_run['dry_run'] = True

    if ctx.invoked_subcommand is None:
        # typer.echo('Welcome to sugar. For usage, try --help.')
        raise typer.Exit()


def map_type_from_string(type_name: str) -> Type[Union[str, int, float, bool]]:
    """
    Return a type object mapped from the type name.

    Parameters
    ----------
    type_name : str
        The string representation of the type.

    Returns
    -------
    type
        The corresponding Python type.
    """
    type_mapping = {
        'str': str,
        'string': str,
        'int': int,
        'integer': int,
        'float': float,
        'bool': bool,
        'boolean': bool,
    }
    return type_mapping.get(type_name, str)


def normalize_string_type(type_name: str) -> str:
    """
    Normalize the user type definition to the correct name.

    Parameters
    ----------
    type_name : str
        The string representation of the type.

    Returns
    -------
    str
        The corresponding makim type name.
    """
    type_mapping = {
        'str': 'str',
        'string': 'str',
        'int': 'int',
        'integer': 'int',
        'float': 'float',
        'bool': 'bool',
        'boolean': 'bool',
        # Add more mappings as needed
    }
    return type_mapping.get(type_name, 'str')


def get_default_value(
    arg_type: str, value: Any
) -> Optional[Union[str, int, float, bool]]:
    """Return the default value regarding its type in a string format."""
    if arg_type == 'bool':
        return False if value is None else bool(value)
    elif arg_type == 'int':
        return int(value) if value is not None else None
    elif arg_type == 'float':
        return float(value) if value is not None else None
    elif arg_type == 'str':
        return str(value) if value is not None else None
    return None


def get_default_value_str(arg_type: str, value: Any) -> str:
    """Return the default value regarding its type in a string format."""
    if arg_type == 'str':
        return f'"{value}"'

    if arg_type == 'bool':
        return 'False'

    return f'{value or 0}'


def create_args_string(args: dict[str, dict[str, str]]) -> str:
    """Return a string for arguments for a function for typer."""
    args_rendered = []

    arg_template = (
        '{arg_name}: Optional[{arg_type}] = typer.Option('
        '{default_value}, '
        '"--{name_flag}", '
        'help="{help_text}"'
        ')'
    )

    for name, spec in args.items():
        name_clean = name.replace('-', '_')
        arg_type = normalize_string_type(spec.get('type', 'str'))
        help_text = spec.get('help', '')
        default_value = 'None'

        if not spec.get('required', False) and not spec.get(
            'interactive', False
        ):
            default_value = spec.get('default', '')
            default_value = get_default_value_str(arg_type, default_value)

        arg_str = arg_template.format(
            **{
                'arg_name': name_clean,
                'arg_type': arg_type,
                'default_value': default_value,
                'name_flag': name,
                'help_text': help_text.replace('\n', '\\n'),
            }
        )

        args_rendered.append(arg_str)

    return ', '.join(args_rendered)


def apply_click_options(
    command_function: Callable[..., Any], options: dict[str, Any]
) -> Callable[..., Any]:
    """
    Apply Click options to a Typer command function.

    Parameters
    ----------
    command_function : callable
        The Typer command function to which options will be applied.
    options : dict
        A dictionary of options to apply.

    Returns
    -------
    callable
        The command function with options applied.
    """
    for opt_name, opt_details in options.items():
        opt_args: dict[
            str, Optional[Union[str, int, float, bool, Type[Any]]]
        ] = {}

        opt_data = cast(Dict[str, str], opt_details)
        opt_type_str = normalize_string_type(opt_data.get('type', 'str'))
        opt_default = get_default_value(opt_type_str, opt_data.get('default'))

        if opt_type_str == 'bool':
            opt_args.update({'is_flag': True})

        opt_args.update(
            {
                'default': None
                if opt_data.get('interactive', False)
                else opt_default,
                'type': map_type_from_string(opt_type_str),
                'help': opt_data.get('help', ''),
                'show_default': True,
            }
        )

        click_option = click.option(
            f'--{opt_name}',
            **opt_args,  # type: ignore[arg-type]
        )
        command_function = click_option(command_function)

    return command_function


def create_dynamic_command(
    ext_name: str,
    typer_group: typer.Typer,
    meta: MetaDocs,
) -> None:
    """
    Dynamically create a Typer command with the specified options.

    Parameters
    ----------
    typer_group : typer.Typer
    name : str
        The command name.
    meta : dict
        the action/command metadata
    """
    name = cast(str, meta.get('name', ''))
    args = cast(Dict[str, Dict[str, str]], meta.get('parameters', {}))
    fn_help = cast(str, meta.get('title', ''))

    args_str = create_args_string(args)
    args_param_list: list[str] = []

    for arg, arg_details in args.items():
        arg_clean = arg.replace('-', '_')
        args_param_list.append(f'{arg}={arg_clean}')

    args_param_str = ','.join(args_param_list)

    decorator = typer_group.command(
        name,
        help=fn_help,
    )

    function_code = f'def dynamic_command({args_str}):\n'

    # handle interactive prompts
    for arg, arg_details in args.items():
        arg_clean = arg.replace('-', '_')

    function_code += f'    sugar = sugar_exts.get("{ext_name}")\n'
    function_code += f'    sugar._cmd_{name}({args_param_str})\n'

    local_vars: dict[str, Any] = {}
    exec(function_code, globals(), local_vars)
    dynamic_command = decorator(local_vars['dynamic_command'])

    # Apply Click options to the Typer command
    if 'args' in args:
        options_data = cast(Dict[str, Dict[str, Any]], args.get('args', {}))
        dynamic_command = apply_click_options(dynamic_command, options_data)


def extract_options_and_cmd_args() -> tuple[list[str], list[str]]:
    """Extract arg `options` and `cmd` from the CLI calling."""
    args = list(sys.argv)
    total_args = len(args)

    if '--options' in args:
        options_sep_idx = args.index('--options')
    else:
        options_sep_idx = None

    if '--cmd' in args:
        cmd_sep_idx = args.index('--cmd')
    else:
        cmd_sep_idx = None

    if options_sep_idx is None and cmd_sep_idx is None:
        return [], []

    # check if --options or --cmd are the last ones in the command line
    first_sep_idx = min(
        [(options_sep_idx or total_args), (cmd_sep_idx or total_args)]
    )
    for sugar_arg in [
        '--verbose',
        '--version',
        '--service-group',
        '--group',
        '--services',
        '--service',
        '--all',
        '--config-file',
    ]:
        if sugar_arg not in args:
            continue

        if first_sep_idx < args.index(sugar_arg):
            print(
                '[EE] The parameters --options/--cmd should be the '
                'last ones in the command line.'
            )
            os._exit(1)

    for ind in range(first_sep_idx, total_args):
        sys.argv.pop(first_sep_idx)

    options_sep_idx = options_sep_idx or total_args
    cmd_sep_idx = cmd_sep_idx or total_args

    if options_sep_idx < cmd_sep_idx:
        options_args = args[options_sep_idx + 1 : cmd_sep_idx]
        cmd_args = args[cmd_sep_idx + 1 :]
    else:
        cmd_args = args[cmd_sep_idx + 1 : options_sep_idx]
        options_args = args[options_sep_idx + 1 :]
    return options_args, cmd_args


def extract_root_config(
    cli_list: list[str] = sys.argv,
) -> dict[str, str | bool]:
    """Extract the root configuration from the CLI."""
    params = cli_list[1:]

    # default values
    sugar_file = '.sugar.yaml'
    group = ''
    dry_run = False
    verbose = False

    try:
        idx = 0
        while idx < len(params):
            arg = params[idx]
            if arg not in CLI_ROOT_FLAGS_VALUES_COUNT:
                break

            if arg == '--file':
                try:
                    sugar_file = params[idx + 1]
                except IndexError:
                    pass
            elif arg == '--group':
                try:
                    group = params[idx + 1]
                except IndexError:
                    pass
            elif arg == '--dry-run':
                dry_run = True
            elif arg == '--verbose':
                verbose = True

            idx += 1 + CLI_ROOT_FLAGS_VALUES_COUNT[arg]
    except Exception:
        red_text = typer.style(
            'The sugar config file was not correctly detected. '
            'Using the default .sugar.yaml.',
            fg=typer.colors.RED,
            bold=True,
        )
        typer.echo(red_text, err=True, color=True)

    return {
        'file': sugar_file,
        'group': group,
        'dry_run': dry_run,
        'verbose': verbose,
    }


def _get_command_from_cli() -> str:
    """
    Get the group and task from CLI.

    This function is based on `CLI_ROOT_FLAGS_VALUES_COUNT`.
    """
    params = sys.argv[1:]
    command = ''

    try:
        idx = 0
        while idx < len(params):
            arg = params[idx]
            if arg not in CLI_ROOT_FLAGS_VALUES_COUNT:
                command = f'flag `{arg}`' if arg.startswith('--') else arg
                break

            idx += 1 + CLI_ROOT_FLAGS_VALUES_COUNT[arg]
    except Exception as e:
        print(e)

    return command


def run_app() -> None:
    """Run the typer app."""
    root_config = extract_root_config()

    config_file_path = cast(str, root_config.get('file', '.sugar.yaml'))

    cli_completion_words = [
        w for w in os.getenv('COMP_WORDS', '').split('\n') if w
    ]

    if not _check_sugar_file(config_file_path) and cli_completion_words:
        # autocomplete call
        root_config = extract_root_config(cli_completion_words)
        config_file_path = cast(str, root_config.get('file', '.sugar.yaml'))
        if not _check_sugar_file(config_file_path):
            return

    for sugar_ext in sugar_exts.values():
        sugar_ext.load(
            file=config_file_path,
            group=cast(str, root_config.get('group', '')),
            dry_run=cast(bool, root_config.get('dry_run', False)),
            verbose=cast(bool, root_config.get('verbose', False)),
        )

    commands: dict[str, list[MetaDocs]] = {}
    actions: list[str] = []

    for ext_name, ext_class in extensions.items():
        ext_obj = ext_class()
        commands[ext_name] = []

        actions = ext_obj.actions

        for action in actions:
            fn_name = f'_cmd_{action}'
            fn = getattr(ext_obj, fn_name)
            title = fn._meta_docs.get('title', '')

            commands[ext_name].append(
                cast(
                    MetaDocs,
                    {
                        'name': action,
                        'title': title,
                        'parameters': cast(
                            MetaDocsParams, fn._meta_docs.get('parameters', {})
                        ),
                    },
                )
            )

    # Add dynamically created commands to Typer app
    for ext_name, actions_meta in commands.items():
        ext_obj = extensions[ext_name]()

        if not ext_obj:
            SugarLogs.raise_error(f'Extension not found ({ext_name}).')

        typer_group = typer.Typer(
            help=ext_obj.__doc__,
            invoke_without_command=True,
        )
        typer_groups[ext_name] = typer_group

        for action_meta in actions_meta:
            create_dynamic_command(ext_name, typer_group, action_meta)

    for ext_name, typer_group in typer_groups.items():
        app.add_typer(typer_group, name=ext_name, rich_help_panel='COMMAND')

    try:
        app()
    except SystemExit as e:
        # code 2 means code not found
        error_code = 2
        if e.code != error_code:
            raise e

        command_used = _get_command_from_cli()
        typer.secho(f'Command {command_used} not found.', fg='red')

        raise e


if __name__ == '__main__':
    run_app()



"""Sugar plugin for docker compose."""

from __future__ import annotations

from sugar.docs import docparams
from sugar.extensions.base import SugarBase
from sugar.logs import SugarErrorType, SugarLogs

doc_group = {
    'group': 'Specify the group name of the services you want to use.'
}
doc_service = {'service': 'Set the service for the container call.'}
doc_services = {
    'services': 'Set the services separated by comma for the container call.'
}
doc_all_services = {'all': 'Use all services for the command.'}
doc_options = {
    'options': (
        'Specify the options for the backend command. E.g.: `--options -d`.'
    )
}
doc_cmd = {
    'cmd': (
        'Specify the COMMAND for some docker-compose command. '
        'E.g.: --cmd python -c print(1).'
    )
}

doc_common_no_services = {
    **doc_group,
    **doc_options,
}

doc_common_service = {
    **doc_group,
    **doc_service,
    **doc_options,
}

doc_common_services = {
    **doc_group,
    **doc_services,
    **doc_all_services,
    **doc_options,
}

doc_common_services_no_options = {
    **doc_group,
    **doc_services,
    **doc_all_services,
}


class SugarCompose(SugarBase):
    """SugarCompose provides the docker compose commands."""

    @docparams(doc_common_service)
    def _cmd_attach(
        self,
        service: str = '',
        options: str = '',
    ) -> None:
        """
        Attach to a service's running container.

        Attach local standard input, output, and error streams to a service's
        running container.

        Note: This is an experimental feature.
        """
        services_names = self._get_service_name(service)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'attach', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_build(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Build or rebuild services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'build', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_config(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Parse, resolve and render compose file in canonical format."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'config', services=services_names, options_args=options_args
        )

    @docparams(doc_common_no_services)
    def _cmd_cp(
        self,
        options: str = '',
    ) -> None:
        """
        Copy files/folders between a services and local filesystem.

        Note: This is an experimental feature.
        """
        options_args = self._get_list_args(options)
        self._call_backend_app('cp', services=[], options_args=options_args)

    @docparams(doc_common_services)
    def _cmd_create(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Create containers for a service."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'create', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_down(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Stop and remove containers, networks."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = [*self._get_list_args(options), *['--remove-orphans']]

        self._call_backend_app(
            'down',
            services=services_names,
            options_args=options_args,
        )

    @docparams(doc_common_services)
    def _cmd_events(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Receive real time events from containers."""
        # port is not complete supported
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'events', services=services_names, options_args=options_args
        )

    @docparams({**doc_common_service, **doc_cmd})
    def _cmd_exec(
        self,
        service: str,
        options: str = '',
        cmd: str = '',
    ) -> None:
        """Execute a command in a running container."""
        services_names = self._get_service_name(service)
        options_args = self._get_list_args(options)
        cmd_args = self._get_list_args(cmd)
        self._call_backend_app(
            'exec',
            services=services_names,
            options_args=options_args,
            cmd_args=cmd_args,
        )

    @docparams(doc_common_services)
    def _cmd_images(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """List images used by the created containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'images', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_kill(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Force stop service containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'kill', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_logs(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """View output from containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'logs', services=services_names, options_args=options_args
        )

    @docparams(doc_common_no_services)
    def _cmd_ls(
        self,
        options: str = '',
    ) -> None:
        """
        List running compose projects.

        Note: This is an experimental feature.
        """
        options_args = self._get_list_args(options)
        self._call_backend_app('ls', services=[], options_args=options_args)

    @docparams(doc_common_services)
    def _cmd_pause(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Pause services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'pause', services=services_names, options_args=options_args
        )

    @docparams(doc_common_no_services)
    def _cmd_port(
        self,
        service: str,
        options: str = '',
    ) -> None:
        """Print the public port for a port binding."""
        # port is not complete supported
        services_names = self._get_service_name(service)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'port', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_ps(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """List containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'ps', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_pull(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Pull service images."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'pull', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_push(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Push service images."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'push', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_restart(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Restart service containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'restart', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_rm(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """
        Remove stopped service containers.

        By default, anonymous volumes attached to containers will not be
        removed. You can override this with -v. To list all volumes, use
        "docker volume ls".

        Any data which is not in a volume will be lost.
        """
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'rm', services=services_names, options_args=options_args
        )

    @docparams({**doc_common_service, **doc_cmd})
    def _cmd_run(
        self,
        service: str,
        options: str = '',
        cmd: str = '',
    ) -> None:
        """Run a one-off command on a service."""
        if not service:
            SugarLogs.raise_error(
                '`run` sub-command expected --service parameter.',
                SugarErrorType.SUGAR_MISSING_PARAMETER,
            )
        services_names = self._get_service_name(service)
        options_args = self._get_list_args(options)
        cmd_args = self._get_list_args(cmd)
        self._call_backend_app(
            'run',
            services=services_names,
            options_args=options_args,
            cmd_args=cmd_args,
        )

    @docparams(doc_common_service)
    def _cmd_scale(
        self,
        service: str = '',
        options: str = '',
    ) -> None:
        """
        Scale services.

        Note: This is an experimental feature.
        """
        options_args = self._get_list_args(options)
        self._call_backend_app('scale', services=[], options_args=options_args)

    @docparams(doc_common_services)
    def _cmd_start(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Start services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'start', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_stop(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Stop services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'stop', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_top(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Display the running processes."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'top', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_unpause(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Unpause services."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'unpause', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_up(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """Create and start containers."""
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'up', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_wait(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """
        Block until the first service container stops.

        Note: This is an experimental feature.
        """
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'wait', services=services_names, options_args=options_args
        )

    @docparams(doc_common_services)
    def _cmd_watch(
        self,
        services: str = '',
        all: bool = False,
        options: str = '',
    ) -> None:
        """
        Watch build context.

        Watch build context for service and rebuild/refresh containers when
        files are updated.

        Note: This is an experimental feature.
        """
        services_names = self._get_services_names(services=services, all=all)
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'watch', services=services_names, options_args=options_args
        )

    @docparams(doc_options)
    def _cmd_version(
        self,
        options: str = '',
    ) -> None:
        """Show the Docker Compose version information."""
        options_args = self._get_list_args(options)
        self._call_backend_app(
            'version', services=[], options_args=options_args
        )



"""Sugar Plugin for Containers Statics."""

from __future__ import annotations

import datetime
import io

from dataclasses import dataclass, field
from itertools import tee
from typing import Any, Iterable

import plotille

from rich.text import Text
from textual.app import App, ComposeResult
from textual.reactive import Reactive
from textual.widget import Widget
from textual.widgets import Header

from sugar.console import get_terminal_size
from sugar.docs import docparams
from sugar.extensions.compose import (
    SugarCompose,
    doc_common_services_no_options,
)
from sugar.inspect import get_container_name, get_container_stats
from sugar.logs import SugarErrorType, SugarLogs

CHART_WINDOW_DURATION = 60
CHART_TIME_INTERVAL = 1


@dataclass
class StatsData:
    """Record stats from containers."""

    times: list[datetime.datetime] = field(default_factory=list)
    mem_usages: list[float] = field(default_factory=list)
    cpu_usages: list[float] = field(default_factory=list)


class StatsPlot:
    """Plot containers statistic data."""

    def __init__(
        self,
        container_names: list[str],
        window_duration: int = CHART_WINDOW_DURATION,
        interval: int = CHART_TIME_INTERVAL,
    ):
        """
        Initialize StatsPlot.

        Parameters
        ----------
        container_names: list
            Names of the Docker containers.
        window_duration: int
            Duration of the window frame for the data in seconds.
        interval: int
            Interval between data points, in seconds.
        """
        self.container_names = container_names
        self.window_duration = window_duration
        self.interval = interval

        self.create_chart()
        self.reset_data()

    def create_chart(self) -> None:
        """Create a new chart."""
        self.fig_mem = plotille.Figure()
        self.fig_cpu = plotille.Figure()

        self.resize_chart()

        self.chart_colors: dict[str, tuple[Iterable[Any], ...]] = {
            'mem': tee(self.fig_mem._color_seq),
            'cpu': tee(self.fig_cpu._color_seq),
        }

        self.stats: dict[str, StatsData] = {
            name: StatsData(times=[], mem_usages=[], cpu_usages=[])
            for name in self.container_names
        }

        for name in self.container_names:
            container_stats = self.stats[name]
            # Add data to plots
            self.fig_mem.plot(
                container_stats.times,
                container_stats.mem_usages,
                label=name,
            )
            self.fig_cpu.plot(
                container_stats.times,
                container_stats.cpu_usages,
                label=name,
            )

    def resize_chart(self) -> None:
        """Resize chart."""
        console_width, console_height = get_terminal_size()

        chart_height = min((console_height - 20) // 2, 10)
        chart_width = console_width - 30

        self.fig_mem.width = chart_width
        self.fig_mem.height = chart_height
        self.fig_cpu.width = chart_width
        self.fig_cpu.height = chart_height

    def reset_chart(self) -> None:
        """Reset chart state."""
        self.fig_mem._plots.clear()
        self.fig_cpu._plots.clear()

        self.fig_mem._color_seq = tee(self.chart_colors['mem'][0])[1]
        self.fig_cpu._color_seq = tee(self.chart_colors['cpu'][0])[1]

        self.resize_chart()

    def reset_data(self) -> None:
        """Generate a clean data."""
        current_time = datetime.datetime.now()

        for name in self.container_names:
            container_stats = self.stats[name]

            container_stats.mem_usages.clear()
            container_stats.cpu_usages.clear()
            container_stats.times.clear()

            container_stats.mem_usages.extend([0.0] * CHART_WINDOW_DURATION)
            container_stats.cpu_usages.extend([0.0] * CHART_WINDOW_DURATION)
            container_stats.times.extend(
                [
                    current_time
                    - datetime.timedelta(seconds=i * CHART_TIME_INTERVAL)
                    for i in range(CHART_WINDOW_DURATION)
                ][::-1]
            )

            # Add data to plots
            self.fig_mem.plot(
                container_stats.times,
                container_stats.mem_usages,
                label=name,
            )
            self.fig_cpu.plot(
                container_stats.times,
                container_stats.cpu_usages,
                label=name,
            )

    def plot_stats(self) -> None:
        """
        Plot containers statistic.

        Plots the memory and CPU usage of multiple Docker containers over
        time in a single chart for each metric.
        """
        current_time = datetime.datetime.now()

        for name in self.container_names:
            mem_usage, cpu_usage = get_container_stats(name)

            # Update and maintain window for stats
            container_stats = self.stats[name]
            container_stats.times.append(current_time)
            container_stats.mem_usages.append(round(mem_usage, 2))
            container_stats.cpu_usages.append(round(cpu_usage, 2))

            container_stats.times = container_stats.times[
                -self.window_duration :
            ]
            container_stats.mem_usages = container_stats.mem_usages[
                -self.window_duration :
            ]
            container_stats.cpu_usages = container_stats.cpu_usages[
                -self.window_duration :
            ]

        self.reset_chart()

        for name in self.container_names:
            container_stats = self.stats[name]
            # Add data to plots
            self.fig_mem.plot(
                container_stats.times,
                container_stats.mem_usages,
                label=name,
            )
            self.fig_cpu.plot(
                container_stats.times,
                container_stats.cpu_usages,
                label=name,
            )


class StatsPlotWidget(Widget):  # type: ignore
    """Plot Docker Stats Widget."""

    content: Reactive[str] = Reactive('')

    DEFAULT_CSS = """
        Plot {
            width: 100%;
            height: 100%;
        }
    """

    def __init__(
        self, container_names: list[str], *args: str, **kwargs: Any
    ) -> None:
        """Initialize StatsPlotWidget."""
        self.container_names = container_names
        super().__init__(*args, **kwargs)

    def on_mount(self) -> None:
        """Set up the widget."""
        # Set up a periodic update, adjust the interval as needed
        interval_time = CHART_TIME_INTERVAL
        self.set_interval(
            interval_time, self.update_plot
        )  # Update every second
        self.stats_plot = StatsPlot(
            container_names=self.container_names,
            window_duration=CHART_WINDOW_DURATION,
            interval=interval_time,
        )

    async def update_plot(self) -> None:
        """Update plot data."""
        self.stats_plot.plot_stats()
        self.content = (
            'Memory Usage (MB):\n'
            + self.stats_plot.fig_mem.show(legend=False)
            + '\n\nCPU Usage (%):\n'
            + self.stats_plot.fig_cpu.show(legend=True)
        )

    def render(self) -> Text:
        """Render the widget."""
        return Text.from_ansi(self.content)


class StatsPlotApp(App[str]):  # type: ignore
    """StatsPlotApp app class."""

    TITLE = 'Sugar Containers Stats'
    container_names: list[str]

    def __init__(
        self, container_names: list[str], *args: str, **kwargs: Any
    ) -> None:
        """Initialize StatsPlotApp."""
        self.container_names = container_names
        super().__init__(*args, **kwargs)

    def compose(self) -> ComposeResult:
        """Compose the app."""
        yield Header()
        yield StatsPlotWidget(self.container_names)


class SugarStats(SugarCompose):
    """SugarStats provides special commands not available on docker-compose."""

    @docparams(doc_common_services_no_options)
    def _cmd_plot(
        self,
        services: str,
        all: bool,
    ) -> None:
        """Call the plot command."""
        _out = io.StringIO()
        _err = io.StringIO()

        services_names = self._get_services_names(services=services, all=all)

        super()._call_backend_app(
            'ps',
            services=services_names,
            options_args=['-q'],
            _out=_out,
            _err=_err,
        )

        if self.dry_run:
            return

        raw_out = _out.getvalue()

        if not raw_out:
            service_names_txt = ', '.join(services_names)
            SugarLogs.raise_error(
                f'No container found for the services: {service_names_txt}',
                SugarErrorType.SUGAR_NO_SERVICES_RUNNING,
            )

        containers_ids = [cids for cids in raw_out.split('\n') if cids]

        containers_names = [get_container_name(cid) for cid in containers_ids]
        app = StatsPlotApp(containers_names)
        app.run()



"""Sugar class for containers."""

from __future__ import annotations

from typing import Optional, Type, cast

from sugar.extensions.base import SugarBase
from sugar.extensions.compose import SugarCompose
from sugar.extensions.compose_ext import SugarComposeExt

try:
    from sugar.extensions.stats import SugarStats
except ImportError:
    # SugarStats is optional (extras=tui)
    SugarStats = cast(Optional[Type[SugarBase]], None)  # type: ignore


extensions: dict[str, Type[SugarBase]] = {
    'compose': SugarCompose,
    'compose-ext': SugarComposeExt,
    **{'stats': SugarStats for i in range(1) if SugarStats is not None},
}
